from .DefDict import DefDict
from .UnitType import UnitType
from .MapRule import MapRule
from .StdUnit import *
import SpeedTest


__author__ = 'UCLA LEMUR'
__credits__ = 'The Laboratory for Embedded Machines and Ubiquitous Robots'
__url__ = 'https://github.com/Suke0811/DefDict/'
__description__ = 'Defined Dictionary'



from defdict import DefDict
from defdict import StdUnit as unit
from typing import Any


class SpaceDefinition:
    default_definition = dict()
    default_dtype = Any
    default_name = None
    default_prefixes = None
    default_suffixes = None
    default_format_rule = None
    default_shape = None
    default_rules = None
    default_nested_def = True

    def __int__(self, definition=None, dtype=None, name=None, prefixes=None, suffixes=None, format_rule=None,
                shape=None,
                rules=None, nested_def=None):
        if definition is None: definition = self.default_definition
        if dtype is None: dtype = self.default_dtype
        if name is None: name = self.default_name
        if prefixes is None: name = self.default_name
        if suffixes is None: suffixes = self.default_suffixes
        if format_rule is None: format_rule = self.default_format_rule
        if shape is None: shape = self.default_shape
        if rules is None: rules = self.default_rules
        if nested_def is None: nested_def = self.default_nested_def

        self.definition = definition
        self.dtype = dtype
        self.name = name
        self.prefixes = prefixes
        self.suffixes = suffixes
        self.format_rule = format_rule
        self.shape = shape
        self.rules = rules
        self.nested_def = nested_def

    def to_dict(self):
        return dict(definition=self.definition, dtype=self.dtype, name=self.name, prefixes=self.prefixes,
                    suffixes=self.suffixes, format_rule=self.format_rule, shape=self.shape,
                    rules=self.rules, nested_def=self.nested_def)

    def to_DefDict(self):
        return DefDict(**self.to_dict())



#
# class Pos2D(SpaceDefinition):
#     definition = dict(x=unit.Pos, y=unit.Pos)
#     name = 'pos'
#
# class Euler(SpaceDefinition):
#     definition = dict(th_x=unit.Ang, th_y=unit.Ang, th_z=unit.Ang)



from defdict.StdUnit import *

import copy
from typing import Any


'''Separator'''
SEPARATOR = '.'
NPS = NON_PREFIX_SEPARATOR = '_'

'''Position'''
POS_2D = dict(x=Pos, y=Pos)
POS_2D_mm = dict(x=Pos(unit='mm'), y=Pos(unit='mm'))
POS_3D = dict(x=Pos, y=Pos, z=Pos)
POS_3D_mm = dict(x=Pos(unit='mm'), y=Pos(unit='mm'), z=Pos(unit='mm'))

'''Rotation'''
QUAT = dict(quat=dict(qw=0.0, qx=1, qy=0.0, qz=0.0))

ROT_MAT = dict(rot_mat=dict(r01=1.0, r02=0.0, r03=0.0,
                  r11=0.0, r12=1.0, r13=0.0,
                  r21=0.0, r22=0.0, r23=1.0,))

AXIS_ANGLE = dict(ax=0.0, ay=0.0, az=1.0, angle=Ang)
AXIS_ANGLE_VEC = dict(ax=0.0, ay=0.0, az=0.0)
EULER_3D = dict(a=float, b=float, c=float)

'''Position velocity'''

VEL_POS_2D = dict(d_x=Vel, d_y=Vel)
VEL_POS_3D = dict(d_x=Vel, d_y=Vel, d_z=Vel)

'''Angular velocity'''
VEL_ROT_2D = dict(d_th=AngVel)
VEL_ROT_3D = dict(d_a=AngVel, d_b=AngVel, d_c=AngVel)

'''Position Acceleration'''
ACC_POS_2D = dict(dd_x=Acc, dd_y=Acc)
ACC_POS_3D = dict(dd_x=Acc, dd_y=Acc, dd_z=Acc)

'''Angular Acceleration'''
ACC_ROT_2D = dict(dd_th=AngAcc)
ACC_ROT_3D = dict(dd_a=AngAcc, dd_b=AngAcc, dd_c=AngAcc)


'''jacobian'''
JACOB_2D = dict(Jb00=float, J02=float, J03=float,
                Jb11=float, J12=float, J13=float,
                Jb21=float, J22=float, J23=float)

JACOB_3D = dict(Jb01=float, J02=float, J03=float, J04=float, J05=float, J06=float,
                Jb11=float, J12=float, J13=float, J14=float, J15=float, J16=float,
                Jb21=float, J22=float, J23=float, J24=float, J25=float, J26=float,
                Jb31=float, J32=float, J33=float, J34=float, J35=float, J36=float,
                Jb41=float, J42=float, J43=float, J44=float, J45=float, J46=float,
                Jb51=float, J52=float, J53=float, J54=float, J55=float, J56=float)


def define(prefix, num, dtype=Any, separater='.', link=False):
    d = {}
    if isinstance(num, list):
        for n in num:
            key = prefix + separater + str(n)
            if link:
                d[key] = copy.copy(dtype)
            else:
                d[key] = copy.deepcopy(dtype)
    else:
        for i in range(num):
            key = prefix + separater + str(i)
            if link:
                d[key] = copy.copy(dtype)
            else:
                d[key] = copy.deepcopy(dtype)
    ret = d
    return ret



from defdict import DefDict, MapRule
from defdict.StdDefinitions import POS_3D, ROT_MAT

import numpy as np
import copy

class Tmat:
    def __init__(self, values, fr_from, fr_to):
        self.tmat = self.create_Tmat()
        self.fr_from = 'base'
        self.fr_tp = 'end'


    def update(self, values):
        if self.is_Tmat(values):
            pass

        if self.is_np_Tmat(values):
            pass





    def transpose(self):
        T = self.as_np()
        return np.transpose(T)

    def as_np(self):
        return self.tmat.as_ruled()


    @staticmethod
    def is_np_Tmat(values):
        is_valid = False
        if isinstance(values, np.ndarray):
            if values.shape == (4, 4):
                is_valid = True
        return is_valid

    @staticmethod
    def is_Tmat(values):
        return isinstance(values, Tmat)


    def create_Tmat(self):
        def T_mat_rule(r11, r12, r13,
                       r21, r22, r23,
                       r31, r32, r33,
                       x, y, z):
            return np.array([[r11, r12, r13, x],
                             [r21, r22, r23, y],
                             [r31, r32, r33, z],
                             [0, 0, 0, 1]])

        def Tmat2dict(row0, row1, row2):
            return dict(r11=row0[0], r12=row0[1], r13=row0[2],
                        r21=row1[0], r22=row1[1], r23=row1[2],
                        r31=row2[0], r32=row2[1], r33=row2[2],
                        x=row0[3], y=row1[3], z=row2[3])

        T_keys = list(ROT_MAT.keys()) + list(POS_3D.keys())
        return DefDict(definition=(ROT_MAT, POS_3D),
                    format_rule=MapRule(T_keys, func=T_mat_rule, inv_func=Tmat2dict, to_list=True))



    def __matmul__(self, other):
        current_np = self.as_np()
        other_np = None
        if self.is_Tmat(other):
            other_np = other.as_np()

        if self.is_np_Tmat(other):
                other_np = other

        if other_np is None:
            raise TypeError(f'{other} must be type of 4*4 numpy array or Tmat')

        return current_np @ other_np



from defdict import DefDict
from typing import Any

class MapRule:
    def __init__(self, origin, func=None, target=None, inv_func=None, with_target=False, to_list=False):
        if origin is None:
            self.origin = None
        else:
            self.origin = DefDict(origin)
        if func is not None and not callable(func):
            raise TypeError(f'func: {func} is not callable')
        self.func = func

        if target is None:
            self.target = None
        else:
            self.target = DefDict(target)
        if inv_func is not None and not callable(inv_func):
            raise TypeError(f'func: {func} is not callable')
        self.inv_func = inv_func
        self.inherit_units = True
        self.to_list = to_list
        self.with_target = with_target

    def map(self, origin, target=None):
        return self._apply_map(self.origin, origin, self.func, self.target, target)

    def inv_map(self, inv_origin, inv_target=None):
        return self._apply_map(self.target, inv_origin, self.inv_func, self.origin, inv_target)

    def _apply_map(self, origin, origin_data, func, target, target_data):
        # if self.target is not None and self.target != name:
        #     return # if not
        if origin is None:
            origin = DefDict(origin_data)
        else:
            if isinstance(origin_data, DefDict) and self.inherit_units:
                new_origin_def = origin.DEF
                for k, o_def in origin_data.DEF.items():
                    if k in origin.keys():
                        if origin._definition[k][0] is Any:
                            new_origin_def[k] = o_def
                origin = DefDict(new_origin_def)
        origin.set(origin_data)

        if target is None:
            if target_data is not None:
                target = DefDict(target_data)
        else:
            if isinstance(target_data, DefDict) and self.inherit_units:
                new_target_def = target.DEF
                for k, t_def in target_data.DEF.items():
                    if k in target.keys():
                        if target._definition[k][0] is Any:
                            new_target_def[k] = t_def
                target = DefDict(new_target_def)

        if func is None:
            if target is None:
                return origin
            else:
                return self._positional_rule(origin, target)
        else:
            if target is not None:
                ret = self._apply_func(origin, func, target)
                if ret is None:
                    return ret
                target.set(ret)
                return target
            else:
                ret = self._apply_func(origin, func, target)
                return ret

    def _apply_func(self, origin, func, target):
        if self.with_target and target is not None:
            if self.to_list:
                ret = func(*origin.list(), *target.list())
            else:
                ret = func(origin, target)
        else:
            if self.to_list:
                ret = func(*origin.list())
            else:
                ret = func(origin)
        return ret

    @staticmethod
    def _positional_rule(origin, target):
        length = min([len(origin), len(target)])
        count = 0

        new_def = {}
        values = []
        for k_o, k_t in zip(origin.keys(), target.keys()):
            if count > length:
                break
            new_def[k_t] = origin._definition[k_o][0]
            values.append(origin._data[k_o][0])
        o = DefDict(new_def).update(values)
        return target.set(o)

    def if_rule_apply(self, keys):
        # true when all bind_from.keys() are in keys
        return all(elem in keys for elem in self.origin.keys())

    def if_rule_out_contains(self, keys):
        # true when all keys are in bind_to.keys()
        if self.target is None:
            return False
        return all(elem in self.target.keys() for elem in keys)



import numpy as np
from defdict import DefDict, UnitType, MapRule
from typing import Union
from scipy.spatial.transform import Rotation as R


"""
Here are predefined physical quantities that are predefined. You can define your own by extending UnitType.
You can also change the settings when you create a instance as follow:
dict(x=Pos(unit='cm', drange=('-1km', '1000m'), dtype=int), y=Pos())
Note that if you do:
Pos_cm = Pos(unit='cm', drange=('-1km', '1000m'), dtype=int)
dict(x=Pos_cm, y=Pos_cm)
changing values in Pos_cm will affect the both
"""


class Example(UnitType):
    # unit
    default_unit = 'dimensionless'
    # default value for this unit, if none then we do dtype(), i.e. float() = 0.0
    default_value = None
    # data type
    default_dtype = float
    # dimension i.e. 1D array of 3 columns -> 3 or 2D array 2 by 3 -> (2, 3)
    default_dim = 1
    # data range for this unit
    default_drange: tuple = (float('-inf'), float('inf'))
    # data mapping to let you create your own unit conversion
    # unit='count'; drange=(0, 4095); drange_map=('0.01m', '1m')
    # This indicates 0 count corresponds to 0.01m and 4096 to 1m and inbetween is linear interpolation
    # drange_map=('-1m', func, '1m') and takes func(val, sv, lv). this let you define custom interpolation
    # you can also have more points. If no func is specified, then it'll be linear interpolation
    # drange=(0, 1024, 2048, 3072 ,4096); drange_map=('10mm', '0.1m', '0.5m', func, '0.75m', ,'1m')
    default_drange_map: tuple = None
    # data scale. (-1, 1) -> -100% to 100%. (0, 1) -> 0% to 100%
    defualt_drange_scale: tuple = (-1, 1)


class Time(UnitType):
    default_unit = 's'
    default_drange = (0, float('inf'))
    defualt_drange_scale = (0, 1)

class Pos(UnitType):
    default_unit = 'm'

class Vel(UnitType):
    default_unit = 'm/s'

class Acc(UnitType):
    default_unit = 'm/s**2'

class Ang(UnitType):
    default_unit = 'rad'

class AngVel(UnitType):
    default_unit = 'rad/s'

class AngAcc(UnitType):
    default_unit = 'rad/s**2'

class Mass(UnitType):
    default_unit = 'kg'
    default_drange = (0, float('inf'))
    defualt_drange_scale = (0, 1)

class Percent(UnitType):
    default_unit = 'percent'
    defualt_drange_scale = (0, 1)

class Count(UnitType):
    default_unit = 'count'
    default_dtype = int
    default_drange = (0, 4095)
    defualt_drange_scale = (0, 1)

class CountSensor(UnitType):
    default_unit = 'count'
    default_dtype = int
    default_drange = (0, 4095)
    default_drange_map = ('0.1m', '1m')
    defualt_drange_scale = (0, 1)

class VelCreate(UnitType):
    default_unit = 'rad/s'
    default_drange = (-5,5)

class CountVel(UnitType):
    default_unit = 'count'
    default_dtype = int
    default_drange = (-500, 500)
    default_drange_map = ('-5rad/s', '5rad/s')
    defualt_drange_scale = (0, 1)

class Float(UnitType):
    defualt_unit = 'dimensionless'
    default_dtype = float

class Int(UnitType):
    defualt_unit = 'dimensionless'
    default_dtype = int



import numpy as np
from defdict import DefDict, UnitType, MapRule
from scipy.spatial.transform import Rotation as R
from defdict.StdUnit import Ang, Pos, Float

# Rotation
# if you do
dict(qw=float, qx=float, qy=float, qz=float)
# this will be automatically set to
dict(qw=Float(drange=(-1, 1)), qx=Float(drange=(0, 1)), qy=Float(drange=(0, 1)), qz=Float(drange=(0, 1)))
# so unitless (there will be no unit dimension check)


quat = dict(definition=dict(qw=float, qx=float, qy=float, qz=float), name='quat')
rot3d = dict(definition=dict(r11=float, r12=float, r13=float,
                                 r21=float, r22=float, r23=float,
                                 r31=float, r32=float, r33=float,), name='rot3d', shape=(3, 3))
rot2d = dict(definition=dict(r11=float, r12=float,
                                 r21=float, r22=float,), name='rot2d', shape=(2, 2))
euler = dict(definition=dict(th_x=Ang, th_y=Ang, th_z=Ang, _coord=str), name='euler')
ax_ang = dict(definition=dict(rx=Ang, ry=Ang, rz=Ang), name='ax_ang')
ax_ang4d = dict(definition=dict(ux=float, uy=float, uz=float, alpha=Ang), name='ax_ang4d')

def from_ax_ang4d(ax_ang4d):
    ax_angle = ax_ang4d[0:3] * ax_ang4d[-1]
    return R.from_rotvec(ax_angle)

def as_ax_ang_4d(r):
    ax_angle = r.as_rotvec()
    norm = np.linalg.norm(ax_angle)
    u_vect = ax_angle / norm
    return np.append(u_vect, norm)

conversion_rules = [
    # from or to quat
    MapRule(quat, lambda quat: R.from_quat(quat.ndarray()).as_matrix(), rot3d,
                  lambda rot3d: R.from_matrix((rot3d.ndarray())).as_quat()),
    MapRule(quat, lambda quat: R.from_quat(quat.ndarray()).as_rotvec(), ax_ang,
                  lambda ax_ang: R.from_rotvec((ax_ang.ndarray())).as_quat()),
    MapRule(quat, lambda quat: as_ax_ang_4d(R.from_quat(quat.ndarray())), ax_ang4d,
                  lambda ax_ang4d: from_ax_ang4d(ax_ang4d.ndarray()).as_quat()),
    MapRule(quat, lambda quat, euler: R.from_quat(quat.ndarray()).as_euler(euler.get('_coord')), euler,
                  lambda euler: R.from_euler(euler.ndarray(), euler.get('_coord')).as_quat(), with_target=True),
    # from or to ax_ang
    MapRule(ax_ang, lambda ax_ang: R.from_rotvec(ax_ang.ndarray()).as_matrix(), rot3d,
                    lambda rot3d: R.from_matrix((rot3d.ndarray())).as_rotvec()),
    MapRule(ax_ang, lambda ax_ang, euler: R.from_rotvec(ax_ang.ndarray()).as_euler(euler.get('_coord')), euler,
                    lambda euler: R.from_euler(euler.ndarray(), euler.get('_coord')).as_rotvec(), with_target=True),
    MapRule(ax_ang, lambda ax_ang: as_ax_ang_4d(R.from_rotvec(ax_ang.ndarray())), ax_ang4d,
                    lambda ax_ang4d: from_ax_ang4d((ax_ang4d.ndarray())).as_rotvec()),
    # from or to rot 3d
    MapRule(rot3d, lambda rot3d, euler: R.from_matrix(rot3d.ndarray()).as_euler(euler.get('_coord')), euler,
                   lambda euler: R.from_euler(euler.ndarray(), euler.get('_coord')).as_rotvec(), with_target=True),
    MapRule(rot3d, lambda rot3d: as_ax_ang_4d(R.from_matrix(rot3d.ndarray())), ax_ang4d,
                   lambda ax_ang4d: from_ax_ang4d((ax_ang4d.ndarray())).as_matrix()),
    # ax_ang4d
    MapRule(ax_ang4d, lambda ax_ang4d, euler: from_ax_ang4d(ax_ang4d.ndarray()).as_euler(euler.get('_coord')), euler,
                      lambda euler: as_ax_ang_4d(R.from_euler(euler.ndarray(), euler.get('_coord'))), with_target=True),
]


class Quaternion(UnitType):
    default_unit = 'quat'
    default_dtype = DefDict(**quat)
    default_dim = 4
    # default_conversion_rules = conversion_rules


class Rotation3DMatix(UnitType):
    default_unit = 'rot3d'
    default_dtype = DefDict(**rot3d)
    default_value = np.eye(3) # identy matrix. Under the hood, it does DefDict.set(default_value)
    default_dim = (3, 3)
   # default_conversion_rules = conversion_rules

class Rotation2DMatix(UnitType):
    default_unit = 'rot2d'
    default_dtype = DefDict(**rot2d)
    default_value = np.eye(2) # identy matrix. Under the hood, it does DefDict.set(default_value)
    default_dim = (2, 2)

class Euler(UnitType):
    default_unit = 'euler'
    # _coord is a protected key (you don't see unless you implicitly access it)
    default_dtype = DefDict(**euler)
    default_value = [0, 0, 0, 'xyz']
    default_dim = 3
    default_conversion_rules = conversion_rules

class AxisAngle(UnitType):
    """
    This is a 3 element version of the Axis-Angle representation.
    The first three elements are unit vector times angle (hence dimension is angle)
    """
    default_unit = 'ax_ang'
    default_dtype = DefDict(**ax_ang)
    default_value = [0, 0, 0] # 0 rotation around z
    default_dim = 3
    default_conversion_rules = conversion_rules

class AxisAngle4D(UnitType):
    """
    This is a 4 element version of the Axis-Angle representation.
    The first three elements are unit vector and the 4th element is angle
    """
    default_unit = 'ax_ang4d'
    default_dtype = DefDict(**ax_ang4d)
    default_value = [0, 0, 1, 0]
    default_dim = 4
    default_conversion_rules = conversion_rules


if __name__ == '__main__':
    import sys
    #sys.setrecursionlimit(30)
    import time
    # st  = time.perf_counter()
    # p = DefDict(dict(x=Pos, quat=Quaternion))
    # print(time.perf_counter()-st)
    # p.set(dict(x=Pos(unit='cm', data=10)))
    # print(time.perf_counter()-st)
    # p.set(dict(quat=[1,0,0,0]))
    # print(time.perf_counter()-st)


    import unyt
    st = time.perf_counter()
    u = unyt.unyt_quantity.from_string('0.1m')
    print('unyt',time.perf_counter()-st)
    st = time.perf_counter()
    d = DefDict(dict(x=float,y=float,z=float))
    print(time.perf_counter()-st)
    # d.format(dict(x=0, y=0)).ndarray()
    # print(time.perf_counter()-st)
    st = time.perf_counter()
    r = DefDict(dict(x=Pos, rot3d=Rotation3DMatix))
    print(time.perf_counter()-st)
    st = time.perf_counter()
    q=Quaternion()
    print('unyt',time.perf_counter()-st)
    q.to(10)
    print('unyt',time.perf_counter()-st)
    pass



from dataclasses import dataclass, field
from typing import Any, Union, Callable, Tuple
import unyt
import numpy as np


Pos = dict(unit='m', dtype=float, drange=(-float('inf'), float('inf')), drange_unit=None, dim=1, default=0.0)
Dist = dict(unit='m', dtype=float, drange=(-float('inf'), float('inf')), drange_unit=None, dim=1, default=0.0)

MIN = 0
MAX = 1

@dataclass
class DataType:
    unit: str = 'dimensionless'
    dtype: type = float
    drange: tuple[(Any, Any)] = field(default_factory=lambda: (-float('inf'), float('inf')))
    drange_map: tuple = None
    dim = 1
    default = None
    scale: tuple[(Any, Any)] = field(default_factory=lambda: (-1, 1))

    def __post_init__(self):
        self._sys_unit = unyt.unyt_quantity.from_string(self.unit)
        drange = []
        for v in self.drange:
            if v is str:    # is str, put it to unyt
                uv = unyt.unyt_quantity.from_string(v)
                if uv.units is unyt.dimensionless:  # in case dimensionless
                    uv *= self._sys_unit

            else:   # otherwise, assume the same unit
                uv = v * self._sys_unit
            drange.append(uv)
        self._sys_drange = tuple(drange)
        self._data = self.get_default()

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, val):
        self.data = val

######################3

    def get_default(self):
        if self.default is None:
            if self.dtype is np.array:
                default = np.zeros(self.dim)
            else:
                try:
                    default = self.dtype()
                except (TypeError, AttributeError):
                    default = 0.0
        return default

    def to_unit(self, val, vdef=None):
        if vdef is None or not issubclass(vdef, DataType):
            nval = val
        else:
            nval = val*vdef._sys_unit.to(self.unit)
        return nval

    def scale_value(self, val):
        """
        drange_unit should be like
        = (val_0, val_1, ...,  val_n)
        or
        = (val_0, func, val_1, val_2, ..., val_n)
        if you skip func, then linear interpolation will be used
        :param val:
        :return:
        """
        val = val * self._sys_unit
        scaled_val = val
        i = iter(self.drange_map)
        sv = next(i)
        while True:
            try:
                nv = next(i)
                if not callable(nv):  # if not callable not function -> this is the next value
                    lv = nv
                    func = lambda val, sv, lv: (val - sv) / (lv - sv)
                else:
                    func = nv
                    lv = next(i)    # if callable, then we need to get next value

                if sv <= val and val <= lv:
                    scaled_val = func(val, sv, lv)
                    break
            except StopIteration:
                break
        return scaled_val

    def to_percent(self, val, vdef=None):
        s_val = None
        if all([vd==float('inf') or vd==-float('inf') for vd in vdef.drange]):
            raise ValueError(f'{val} cannot converted to % because data range is {vdef.drange}')
        else:
#            if
            u_val = val*vdef._sys_unit  # add unit
            p_val = (u_val - self._sys_drange[MIN]) / (self._sys_drange[MAX] - self._sys_drange[MIN])
            s_val = p_val * (self.scale[MAX] - self.scale[MIN]) - self.scale[MIN]
        return s_val


    def to_unit(self, val, v_def=None):
        if v_def is None:
            v_def = self
        unit_val = (val * v_def._sys_unit).to(self._sys_unit)
        return self.enforce_type(unit_val)


    def to_count(self):
        pass

    def enforce_type(self, val):
        #enforced_val =
        if self.dtype is np.array:
            enforced_val = val
        return









@dataclass
class DataFormatter:
    """
    DataFormat for
    """
    def __init__(self, unit: str = 'dimensionless', dtype: type = float,
                 drange: tuple[(Any, Any)] = None, drange_unit: tuple = None, dim=None, default=None):
        self._dtype = dtype
        self._unit = unit
        if drange is None:
            drange = (-float('inf'), float('inf'))
        self._drange = drange
        self._drange_unit = drange_unit
        if dim is None:
            try:
                dim = np.dim(self.dtype())
            except (TypeError, AttributeError):
                dim = 1
        self._dim = dim
        if default is None:
            try:
                default = self.dtype()
            except (TypeError, AttributeError):
                default = float()
        self._default = default




    def to_data(self, val, vdef=None):
        if vdef is None or not issubclass(vdef, DataFormatter):
            nval = val
        else:
            nval = val*vdef.unit.to(self.unit)
        return nval

    def scale_value(self, val):
        """
        drange_unit should be like
        = (val_0, val_1, ...,  val_n)
        or
        = (val_0, func, val_1, val_2, ..., val_n)
        if you skip func, then linear interpolation will be used
        :param val:
        :return:
        """
        val = val * self.unit
        scaled_val = val
        i = iter(self.drange_unit)
        sv = next(i)
        while True:
            try:
                nv = next(i)
                if not callable(nv):  # if not callable not function -> this is the next value
                    lv = nv
                    func = lambda val, sv, lv: (val - sv) / (lv - sv)
                else:
                    func = nv
                    lv = next(i)    # if callable, then we need to get next value

                if sv <= val and val <= lv:
                    scaled_val = func(val, sv, lv)
                    break
            except StopIteration:
                break
        return scaled_val

    def to_percent(self, val, vdef):
        if all([vd==float('inf') or vd==-float('inf') for vd in vdef.drange]):
            raise ValueError(f'{val} cannot converted to % because data range is {vdef.drange}')

    def to_unit(self, val, v_def=None):
        if v_def is None:
            v_def = self
        unit_val = (val * v_def.unit).to(self.unit)
        return self.enforce_type(unit_val)


    def to_count(self):
        pass

    def enforce_type(self, val):
#        enforced_val =
        if self.dtype is np.array:
            enforced_val = val
        return


class UnitType:
    default_unit = 'm'
    default_dtype = float
    default_drange: tuple[(Union[int, float, str], Union[int, float, str])] = ('-1m', '1m')
    default_drange_map: tuple = None
    default_dim = 1
    default_value = None
    defualt_drange_scale: tuple[(Union[int, float], Union[int, float])] = (-1, 1)

    def __init__(self, unit=default_unit, drange=None, default=default_value, dim=None, drange_map=None, drange_scale=None, dtype=default_dtype, ):
        if drange is None:
            drange = self.default_drange
        if dim is None:
            dim = self.default_dim
        if drange_map is None:
            drange_map = self.default_drange_map
        if drange_scale is None:
            drange_scale = self.defualt_drange_scale


    @property
    def unit(self):
        return self._unit
    @unit.setter
    def unit(self, val:str):
        self._unit = unyt.unyt_quantity.from_string(val)

    @property
    def dim(self):
        return self._dim
    @dim.setter
    def dim(self, v):
        self._dim = v

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, v):
        try:
            self._default = self.dtype(v)
        except (TypeError, AttributeError):
            self._default = v

    @property
    def drange_unit(self):
        if self.drange_unit is None:
            self._drange_unit = (self.unit, self.unit)
        return self._drange_unit

    @drange_unit.setter
    def drange_unit(self, v):
        self._drange_unit = v

    @property
    def drange(self):
        return self._drange

    @drange.setter
    def drange(self, v):
        self._drange = v

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, v):
        self._dtype = v


    def to_unit(self, val, v_def=None):
        if v_def is None:
            v_def = self
        unit_val = (val * v_def._sys_unit).to(self._sys_unit)
        return self.enforce_type(unit_val)

    unit: str = 'dimensionless'
    dtype: type = float
    drange: tuple[(Any, Any)] = field(default_factory=lambda: (-float('inf'), float('inf')))


    def __post_init__(self):
        self._sys_unit = unyt.unyt_quantity.from_string(self.unit)
        drange = []
        for v in self.drange:
            if v is str:  # is str, put it to unyt
                uv = unyt.unyt_quantity.from_string(v)
                if uv.units is unyt.dimensionless:  # in case dimensionless
                    uv *= self._sys_unit

            else:  # otherwise, assume the same unit
                uv = v * self._sys_unit
            drange.append(uv)
        self._sys_drange = tuple(drange)
        self._data = self.get_default()

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, val):
        self.data = val

    ######################3

    def get_default(self):
        if self.default is None:
            if self.dtype is np.array:
                default = np.zeros(self.dim)
            else:
                try:
                    default = self.dtype()
                except (TypeError, AttributeError):
                    default = 0.0
        return default

    def to_unit(self, val, vdef=None):
        if vdef is None or not issubclass(vdef, DataType):
            nval = val
        else:
            nval = val * vdef._sys_unit.to(self.unit)
        return nval

    def scale_value(self, val):
        """
        drange_unit should be like
        = (val_0, val_1, ...,  val_n)
        or
        = (val_0, func, val_1, val_2, ..., val_n)
        if you skip func, then linear interpolation will be used
        :param val:
        :return:
        """
        val = val * self._sys_unit
        scaled_val = val
        i = iter(self.drange_map)
        sv = next(i)
        while True:
            try:
                nv = next(i)
                if not callable(nv):  # if not callable not function -> this is the next value
                    lv = nv
                    func = lambda val, sv, lv: (val - sv) / (lv - sv)
                else:
                    func = nv
                    lv = next(i)  # if callable, then we need to get next value

                if sv <= val and val <= lv:
                    scaled_val = func(val, sv, lv)
                    break
            except StopIteration:
                break
        return scaled_val

    def to_percent(self, val, vdef=None):
        s_val = None
        if all([vd == float('inf') or vd == -float('inf') for vd in vdef.drange]):
            raise ValueError(f'{val} cannot converted to % because data range is {vdef.drange}')
        else:
            #            if
            u_val = val * vdef._sys_unit  # add unit
            p_val = (u_val - self._sys_drange[MIN]) / (self._sys_drange[MAX] - self._sys_drange[MIN])
            s_val = p_val * (self.scale[MAX] - self.scale[MIN]) - self.scale[MIN]
        return s_val

    def to_unit(self, val, v_def=None):
        if v_def is None:
            v_def = self
        unit_val = (val * v_def._sys_unit).to(self._sys_unit)
        return self.enforce_type(unit_val)

    def to_count(self):
        pass

    def enforce_type(self, val):
        # enforced_val =
        if self.dtype is np.array:
            enforced_val = val
        return

pass





import copy
import logging
import inspect
import numpy as np
from typing import Any
from UnitType import UnitType
SEPARATOR = '.'

class DefDict:
    reserved = ['get', 'set', 'keys', 'list', 'list_keys', 'ndtall']
    def __init__(self, definition=None, dtype=Any, name=None, prefixes=True, suffixes=True, format_rule=None, shape=None, rules=None, nested_def=True):
        self._definition = dict()
        self._data = dict()
        self._name = name
        self.format_rule = format_rule
        self.shape = shape
        self.rules = []
        self.suffixes = []
        self.prefixes = []
        if definition is not None:
            self.add_def(definition, dtype=dtype, prefixes=prefixes, suffixes=suffixes, format_rule=format_rule,
                         shape=shape, rules=rules, nested_def=nested_def)
    
    def add_def(self, definition, dtype=Any, prefixes=True, suffixes=True, format_rule=None, shape=None, rules=None, nested_def=True):
        if format_rule is not None:
            self.format_rule = format_rule
        if shape is not None:
            self.shape = shape
        if not isinstance(rules, list) and rules is not None:
            rules = [rules]
        if rules is not None:
            self.rules.extend(rules)

        suff_add = []

        if isinstance(definition, tuple):
            for d in definition:  # add as many definition as you want
                suf = self._set_defs(d, dtype, nested_def)
                if suffixes is not False:
                    suff_add.extend(suf)
                if isinstance(d, DefDict):
                    if suffixes is not False:
                        suff_add.extend(d.list_keys())
                    if d.rules is not None:
                        self.rules.extend(d.rules)
        else:
            suf = self._set_defs(definition, dtype, nested_def)
            if suffixes is not False:
                suff_add.extend(suf)

        if isinstance(suffixes, dict):
            suff_add.extend(list(suffixes.keys()))
        if isinstance(suffixes, list):
            suff_add.extend(suffixes)

        if prefixes is True:
            prefixes = self._find_all_prefixes(self.keys())

        if prefixes is not None:
            if isinstance(prefixes, dict):
                prefixes = list(prefixes.keys())
            self._add_prefixes(prefixes)
        if suff_add:
            self._add_suffixes(suff_add)
        return self

    def _find_all_prefixes(self, data_list):
        prefix = {}
        for d in data_list:
            if not isinstance(d, str):
                d = str(d)
            if d.find(SEPARATOR) >= 1:
                prefix.setdefault(d.split(SEPARATOR)[0], d.split(SEPARATOR)[0])
        return list(prefix.keys())


    def ndarray(self, reshape: tuple = None):
        data_list = self.list()
        if reshape is None:
            reshape = self.shape
        return np.array(data_list).reshape(reshape)

    def ndtall(self):
        return self.ndarray((len(self._data), 1))

    def list(self):
        return list(self.values())

    def list_keys(self):
        return list(self._data.keys())

    def get(self, key=None):
        if key is None:
            ret = self.list()[0]    # get the first element when no key has been specified
        else:
            ret = self._data[key][0]
        return ret

    def dict(self):
        ret = {}
        for k, v in self.data.items():
            if isinstance(v, DefDict):
                ret[k] = v.dict()
            else:
                ret[k] = v
        return ret  # return dictionary

    def keys(self, to_int=False):
        if to_int:
            return map(int, self.list_keys())
        else:
            return self.list_keys()

    def as_ruled(self):
        if self.format_rule is None:
            return self.data
        return self.format_rule.map(self.data)

    def remove_prefix(self, prefix=None):
        d = copy.deepcopy(self)
        d.clear()
        for k in self.keys():
            k_seq = k.split(SEPARATOR)
            if len(k_seq) <= 1:
                continue
            if prefix is None:
                k_seq.pop(0)
                d._set_defs({SEPARATOR.join(k_seq): self.DEF[k]})
                d._data[SEPARATOR.join(k_seq)] = self._data.get(k)
            else:
                if prefix in k_seq:
                    k_seq.remove(prefix)
                    d._set_defs({SEPARATOR.join(k_seq): self.DEF[k]})
                    d._data[SEPARATOR.join(k_seq)] = self._data.get(k)

        return d

    def _add_prefixes(self, prefixes):
        if not isinstance(prefixes, list):
            prefixes = [prefixes]
        for name in prefixes:
            self._add_prefix(name)

    def _add_prefix(self, name):
        if name in self.reserved:
            raise AttributeError(f'the prefix {name} is reserved for DefDict')
        if name in self.prefixes:
            logging.debug(f'the prefix {name} is already registered')
            return
        self.prefixes.append(name)

        def method(self, ids=None):
            if ids is None:
                return self.remove_prefix(name)
            elif not isinstance(ids, list):
                ids = [ids]
            ids = list(map(str, ids))
            return self.remove_prefix(name).filter(ids)
        setattr(self, name, method.__get__(self))

    def _add_suffixes(self, suffixes):
        if not isinstance(suffixes, list):
            suffixes = [suffixes]
        for name in suffixes:
            if name is None:
                continue
            self._add_suffix(name)

    def _add_suffix(self, name):
        if name in self.suffixes:
            logging.debug('the suffix is already registered')
            return
        self.suffixes.append(name)
        def method(self):
            d = copy.deepcopy(self)
            d.clear()
            for k, v in self.items():
                if isinstance(v, DefDict):
                    d._set_defs({k: v.DEF[name]})
                    d._data[k] = v._data.get(name)
                elif isinstance(v, dict):
                    d._set_defs({k: v[name]})
                    d._data[k] = v.get(name)
                elif k == name:
                    d._set_defs({k: self.DEF[k]})
                    d._data[k] = v
            return d
        setattr(self, name, method.__get__(self))

    @property
    def data(self):
        ret = {}
        for k, v in self._data.items():
            if not k.startswith('_'):  # protected keys are not accessible normally
                ret[k] = v[0]
        return ret

    @data.setter
    def data(self, ndata):
        self.set(ndata)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    def set(self, ndata, apply_rule=True):
        if ndata is None:
            return self
        if apply_rule:
            if self.rules is not None:
                ret = self._apply_rules(ndata)
                if ret is not None:
                    ndata = ret
            if self.format_rule is not None:
                try:
                    ndata = self.format_rule.inv_bind(ndata)
                except:
                    pass

        if isinstance(ndata, DefDict):
            self._from_defdict(ndata)
        elif isinstance(ndata, dict):
            self._dict2dict(ndata)
        elif isinstance(ndata, list):
            self._list2dict(ndata)
        elif isinstance(ndata, tuple):
            ndata = list(ndata)
            self._list2dict(ndata)
        elif isinstance(ndata, np.ndarray):
            self._list2dict(ndata)
        else:
            self._list2dict([ndata])
        return self

    def set_positional(self, ndata):
        if isinstance(ndata, DefDict):
            ndata = self._positional_rule(ndata, self)
        elif isinstance(ndata, dict):
            ndata = list(ndata.values())
        return self.set(ndata)

    @staticmethod
    def _positional_rule(origin, target):
        length = min([len(origin), len(target)])
        count = 0

        new_def = {}
        values = []
        for k_o, k_t in zip(origin.keys(), target.keys()):
            if count > length:
                break
            new_def[k_t] = origin._definition[k_o][0]
            values.append(origin._data[k_o][0])
        o = DefDict(new_def).update(values)
        return target.set(o)

    def _apply_rules(self, data):
        ret = None
        ret_defdict = None
        if isinstance(data, dict) or isinstance(data, DefDict):
            for rule in self.rules:
                origin = rule.origin
                if origin is None:
                    continue
                keys = origin.list_keys()
                if all(elem in list(data.keys()) for elem in keys):
                    ret = rule.map(origin=data, target=self)
                    if ret is not None:
                        self.set(ret, apply_rule=False)
                        continue

        return ret_defdict

    def set_rule(self, rule):
        if self.rules is None:
            self.rules = []
        if not isinstance(rule, list):
            rule = [rule]
        self.rules.extend(rule)
        return self

    def clear_rules(self):
        self.rules = None
        return self

    def set_name(self, name):
        self._name = name
        return self

    def format(self, ndata):
        """format dose not change the stored data, but format the input into current definition"""
        d = copy.deepcopy(self)
        return d.set(ndata)

    @property
    def DEF(self):
        ret = {}
        for k, v in self._definition.items():
            #if not k.startswith('_'):  # protected keys are not accessible normally
            ret[k] = v[0]
        return ret


    def get_DEF(self, keys):
        key_lists = []
        if isinstance(keys, DefDict):
            key_lists = keys.list_keys()
        elif isinstance(keys, dict):
            key_lists = list(keys.keys())
        elif isinstance(keys, tuple):
            k, v = keys
            key_lists = [k]

        DEFs = []
        for k in key_lists:
            if k in self.DEF.keys():
                DEFs.append(self.DEF[k])
            else:
                raise (f'{k} is not in definition')
        return DEFs

    
    def _set_defs(self, ndef, dtype=Any, nested_dict=False):
        keys = []
        suffixes = []

        if inspect.isclass(dtype) and issubclass(dtype, UnitType):
            dtype = dtype()

        if isinstance(dtype, dict) and nested_dict:
            dtype = DefDict(dtype)
            suffixes.extend(dtype.list_keys())

        if isinstance(ndef, DefDict):
            ndef = ndef.DEF # if DefDict is hand over, get definition dict

        if isinstance(ndef, dict):
            for k, v in ndef.items():
                if isinstance(v, dict) and nested_dict:
                    v = DefDict(v)
                    suffixes.extend(v.list_keys())

                if inspect.isclass(v) and issubclass(v, UnitType):
                    self._definition[k] = [v()]
                    self._data[k] = [v().default]
                elif isinstance(v, UnitType):
                    self._definition[k] = [v]
                    self._data[k] = [v.default]
                elif isinstance(v, DefDict):
                    self._definition[k] = [v]
                    suffixes.extend(v.list_keys())
                   # keys.append(k)
                    self._data[k] = [v]
                elif isinstance(v, type) or v is Any:
                    self._definition[k] = [v]
                    keys.append(k)
                    self._data[k] = [0.0]
                else:
                    self._definition[k] = [type(v)]
                    self._data[k] = [v]

        elif isinstance(ndef, list):
            self._definition.update(dict.fromkeys(ndef, [dtype]))
            keys.extend(ndef)
        elif isinstance(ndef, str):
            self._definition[ndef] = [dtype]
            keys.append(ndef)
        else:
            raise TypeError('You can only add str, dict, or list')
        self.init_data(keys)
        return suffixes



    def init_data(self, keys):
        for k, v in self.DEF.items():
            if k in keys:
                if isinstance(v, UnitType):
                    self._data[k] = [v.default]
                elif v is Any:
                    self._data[k] = [float()] # what should be the init value?
                elif isinstance(v, type):
                    self._data[k] = [v()]
                else:
                    self._data[k] = [v]   # maybe a different way of initialization?
        return self

    def to_default(self, keys=None):
        if keys is None:
            keys = self.list_keys()
        else:
            try:
                iter(keys)
            except:
                keys = [keys]
        return self.init_data(keys)

    def _from_defdict(self, data):
        for k, v in data.items():
            if k in self._data.keys():
                if isinstance(self.DEF[k], DefDict):
                    self._data[k][0].set(v)
                else:
                    self._data[k][0] = self._enforce_type(self.DEF[k], v, data.DEF[k])

    def _dict2dict(self, data: dict):
        for k, v in data.items():
            if k in self._data.keys():
                if self.DEF[k] is DefDict:
                    self._data[k][0].set(v)
                else:
                    self._data[k][0] = self._enforce_type(self.DEF[k], v)

    def _list2dict(self, data):
        length = min(len(data), len(self._definition)) - 1
        for i, key in enumerate(self._data.keys()):
            if i > length:
                break
            if key.startswith('_'):
                continue
            if isinstance(self.DEF[key], DefDict):
                self._data[key][0].set(data[i])
            else:
                self._data[key][0] = self._enforce_type(self.DEF[key], data[i])

    def _check_if_iterable(self, item):
        try:
            iter(item)
            return True
        except TypeError:
            return False

    
    def _enforce_type(self, d_type, value, vdef=None):
        if value is None:     # None will bypass the enforcement
            ret = None
        elif isinstance(d_type, UnitType):
            ret = self._unit_type(d_type, value, vdef)
        elif isinstance(d_type, DefDict):
            ret = d_type.set(value)
        elif d_type is Any:   # If the type is Any, no enforcement
            ret = value
        else:
            try:
                ret = d_type(value)
            except (TypeError, AttributeError):
                ret = value
            # if not isinstance(ret, d_type):
            #     raise TypeError(f'{ret} is not type of {d_type}')
        return ret    # Enforce type in the corresponding definition

    def _unit_type(self, dtype, value, vdef=None):
        if not isinstance(vdef, UnitType):
            vdef = dtype
        return dtype.enforce(value, vdef)

    def bind(self, bind_rule, val=None):
        if val is None:
            val = self.dict()
        self.set(bind_rule.map(val))
        return self

    def inv_bind(self, bind_rule, val=None):
        if val is None:
            val = self.dict()
        self.set(bind_rule.inv_map(val))
        return self

    def assert_data(self, data=None):
        if data is None:
            data = self._data
        else:
            data = dict(data)
        assert (len(data) == len(self.DEF))
        assert (all(type(x) == y for x, y in zip(data.values(), self.DEF.values())))

    def _to_key_list(self, keys):
        if isinstance(keys, dict):
            keys = list(keys.keys())
        if isinstance(keys, str):
            keys = [keys]
        if isinstance(keys, int):
            keys = [str(keys)]
        if not isinstance(keys, list):
            raise TypeError('keys must be either string, int, list, or dict')
        return keys

    def filter(self, keys): #ToDo support all iteratibe
        key_list = []
        if isinstance(keys, tuple):
            for t in keys:
                key_list.extend(self._to_key_list(t))
        else:
            key_list.extend(self._to_key_list(keys))

        key_list = list(map(str, key_list))
        d = copy.deepcopy(self)
        # for k in self.DEF:
        #     if k not in key_list:
        #         d.remove(k)
        # return d    #DefDict

        d.clear()
        for k in key_list:
            if k in self.keys():
                d._set_defs({k: self.DEF[k]})
                d._data[k] = self._data.get(k)
        return d  # DefDict

    def remove(self, key):
        self._data.pop(key)
        self._definition.pop(key)

    def filter_data(self, data):
        if isinstance(data, dict):
            data = list(data.values())
        if not isinstance(data, list):
            data = [data]

        found = []
        for i, value in enumerate(self.list()):
            for search_for in data:
                if value == search_for:
                    found.append(i)
        keys = list(self.keys())
        vals = list(self.values())
        d = copy.deepcopy(self)
        d.clear()
        for index in found:
            d._set_defs({keys[index]: vals[index]})
            d._data[keys[index]] = vals[index]
        return d

    def flat_list(self):
        ret = []
        for elem in self.values():
            try:
                i = iter(elem)
                for e in elem:
                    ret.append(e)
            except TypeError:
                ret.append(elem)
        return ret

    def to_float(self):
        for k, v in self.items():
            if isinstance(v, np.ndarray):
                v.astype(float)
            elif isinstance(v, DefDict):
                self._data[k][0] = v.to_float()
            else:
                try:
                    if v is not None:
                        self._data[k][0] = float(v)
                except ValueError:
                    pass
        return self

    def to_int(self):
        for k, v in self.items():
            if isinstance(v, np.ndarray):
                v.astype(int)
            elif isinstance(v, DefDict):
                self._data[k][0] = v.to_int()
            else:
                try:
                    if v is not None:
                        self._data[k][0] = int(v)
                except ValueError:
                    pass
        return self

    def flatten(self):
        d = dict()
        for k, v in self.items():
            if isinstance(v, DefDict):
                for kk, vv in v.items():
                    d.update({k + SEPARATOR + kk: vv})
            else:
                d.update({k: v})
        return DefDict(d)
        pass


###################################################################
    #dictionary methods
    def clear(self):  # real signature unknown; restored from __doc__
        """ D.clear() -> None.  Remove all items from D. """
        self._data.clear()
        self._definition.clear()

    def copy(self):  # real signature unknown; restored from __doc__
        """ D.copy() -> a shallow copy of D """
        return copy.copy(self)

    def clone(self):
        """
        Deep copy of itself
        """
        return copy.deepcopy(self)

    def items(self):  # real signature unknown; restored from __doc__
        """ D.items() -> a set-like object providing a view on D's items """
        return self.data.items()

    def pop(self, keys, d=None):  # real signature unknown; restored from __doc__
        """
        D.pop(k[,d]) -> v, remove specified key and return the corresponding value.

        If key is not found, default is returned if given, otherwise KeyError is raised
        """
        rets = []
        if not isinstance(keys, list):
            k = [keys]
        for k in keys:
            if k in self._data.keys():
                rets.append(self._data.get(k)[0])
                self.init_data(k)
            else:
                rets.append(d)
        return rets

    def popitem(self, *args, **kwargs):  # real signature unknown
        """
        Remove and return a (key, value) pair as a 2-tuple.

        Pairs are returned in LIFO (last-in, first-out) order.
        Raises KeyError if the dict is empty.
        """
        keys = [k for k, v in args]
        self.pop(keys)

    def setdefault(self, ndata):  # real signature unknown
        """
        Insert key with a value of default if key is not in the dictionary.
        Return the value for key if key is in the dictionary, else default.
        """
        return self.set(ndata)

    def update(self, ndata):  # known special case of dict.update
        """
        same as set
        """
        return self.set(ndata)

    def values(self):  # real signature unknown; restored from __doc__
        """ D.values() -> an object providing a view on D's values """
        return self.data.values()

    @staticmethod
    def fromkeys(keys, value):
        return DefDict(keys, dtype=value)


    def __setitem__(self, key, value):
        if key in self.DEF.keys():
            if isinstance(self._data[key][0], DefDict):  # for nested def dict
                self._data[key][0].set(value)
            else:
                self._data[key][0] = value
        else:
            raise KeyError(f'Key {key} is not in definition')

    def __getitem__(self, item):
        return self._data.__getitem__(item)[0]

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == '_definition':
                d = {dk: dv for dk,dv in self._definition.items()}
                setattr(result, k, d)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result




##############################################################
    # magic methods
    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return {k: v[0].__str__() for k, v in self._data.items()}.__str__()

    def __str__(self):
        return {k: v[0].__str__() for k, v in self._data.items()}.__str__()

##############################################################
    # Math operations
    def _math(self, other, func, immutable=True):
        if immutable:
            current = self.clone()
        else:
            current = self
        if np.isscalar(other):  # if scalar, then add the value to all elements
            for k in current._data.keys():
                try:
                    ret = func(current._data[k][0], other)
                except TypeError:
                    ret = current._data[k][0]
                current._data[k][0] = ret
        else:  # other wise element wise
            if not isinstance(other, DefDict):
                # if not DefDict, create one assuming
                other_defdict = self.clone()
                other_defdict.init_data(other_defdict.list_keys())
                other_defdict.set(other)
            else:
                other_defdict = other
            # sum for corresponding keys
            for k, o in zip(current.filter(other_defdict.keys()).list_keys(),
                        other_defdict.filter(other_defdict.list_keys()).list()):
                try:
                    ret = func(current._data[k][0],  current._enforce_type(current.DEF[k], o, other_defdict.DEF[k]))
                except TypeError:
                    ret = current._data[k][0]
                current._data[k][0] = current._enforce_type(current.DEF[k], ret)
        return current

    def __iter__(self):
        return iter(self.values())

    def __add__(self, other):
        return self._math(other, lambda v, o: v + o, immutable=True)

    def __iadd__(self, other):
        return self._math(other, lambda v, o: v + o, immutable=False)

    def __sub__(self, other):
        return self._math(other, lambda v, o: v - o, immutable=True)

    def __isub__(self, other):
        return self._math(other, lambda v, o: v - o, immutable=False)

    def __mul__(self, other):
        return self._math(other, lambda v, o: v * o, immutable=True)

    def __imul__(self, other):
        return self._math(other, lambda v, o: v * o, immutable=False)

    def __floordiv__(self, other):
        return self._math(other, lambda v, o: v // o, immutable=True)

    def __ifloordiv__(self, other):
        return self._math(other, lambda v, o: v // o, immutable=False)

    def __truediv__(self, other):
        return self._math(other, lambda v, o: v / o, immutable=True)

    def __itruediv__(self, other):
        return self._math(other, lambda v, o: v / o, immutable=False)

    def __mod__(self, other):
        return self._math(other, lambda v, o: v % o, immutable=True)

    def __imod__(self, other):
        return self._math(other, lambda v, o: v % o, immutable=True)

    def __pow__(self, other, modulo=None):  #Todo: modulo implementation
        return self._math(other, lambda v, o: v ** o, immutable=True)

    def __ipow__(self, other, modulo=None):  #Todo: modulo implementation
        return self._math(other, lambda v, o: v ** o, immutable=False)

#############################
    def _compare(self, other, func):
        ret = []
        current = self
        if np.isscalar(other):  # if scalar, then compare the value to all elements
            for k in current._data.keys():
                ret.append(func(current._data[k][0], other))
        else:  # other wise element wise
            if not isinstance(other, DefDict):
                # if not DefDict, create one assuming
                other_defdict = self.clone()
                other_defdict.init_data(other_defdict.list_keys())
                other_defdict.set(other)
            else:
                other_defdict = other
            # sum for corresponding keys
            for k, o in zip(current.filter(other_defdict.keys()).list_keys(),
                        other_defdict.filter(other_defdict.list_keys()).list()):
                ret.append(func(current._data[k][0], o))
        return ret

    def __lt__(self, other):
        return self._compare(other, lambda v, o: v < o)

    def __le__(self, other):
        return self._compare(other, lambda v, o: v <= o)

    def __eq__(self, other):
        return self._compare(other, lambda v, o: v == o)

    def __ne__(self, other):
        return self._compare(other, lambda v, o: v != o)

    def __ge__(self, other):
        return self._compare(other, lambda v, o: v >= o)

    def __gt__(self, other):
        return self._compare(other, lambda v, o: v > o)

    def __round__(self, n=None):
        d = self.clone()
        for k, v in d.items():
            d._data[k][0] = round(v, n)
        return d


if __name__ == '__main__':
    d =DefDict({'leg.0':DefDict({'j.0':1, 'j.1':2},),'leg.1':DefDict({'j.0':1, 'j.1':2},)})
    print(d.flatten())
    print(d)



from defdict import DefDict
from defdict.StdUnit import Pos
from time import perf_counter

st =perf_counter()
d = DefDict(dict(x=float,y=float,z=float))
print(perf_counter()-st)

st =perf_counter()
n = DefDict(dict(x=float,y=float,z=dict(a=32,b=6)))
print(perf_counter()-st)

st =perf_counter()
c = DefDict(dict(x=Pos,y=Pos,z=Pos)).list()
n.set([1,2, [3,4]])
print(perf_counter()-st)



from typing import Any, Union, Tuple
import unyt
from unyt.exceptions import UnitParseError
import numpy as np



Pos = dict(unit='m', dtype=float, drange=(-float('inf'), float('inf')), drange_unit=None, dim=1, default=0.0)
Dist = dict(unit='m', dtype=float, drange=(-float('inf'), float('inf')), drange_unit=None, dim=1, default=0.0)

RESERVED = ['quat', 'rot3d', 'rot2d', 'euler', 'ax_ang', 'ax_ang4d']

MAX=1
MIN=0

class InvalidUnitChangeError(Exception):
    def __init__(self, old_unit, old_dim, new_unit, new_dim):
        self.old_unit = old_unit
        self.old_dim = old_dim
        self.new_unit = new_unit
        self.new_dim = new_dim

    def __str__(self):
        return f"Unit cannot be changed to {self.new_unit} from {self.old_unit} because the dimensions is inconsistent between default {self.old_dim} and new {self.new_dim}"

class UnitType:
    default_unit = 'dimensionless'
    default_dtype = float
    default_dim = 1
    default_value = None
    default_drange = (float('-inf'), float('inf'))
    default_drange_map = None
    defualt_drange_scale = (-1, 1)
    default_conversion_rules: list = []

    def __init__(self, unit=None, drange=None, default=None, dim=None, mapping=None, scale=None, dtype=None, strict=False, data=None):
        if unit is None:
            unit = self.default_unit
        if drange is None:
            drange = self.default_drange
        if dim is None:
            dim = self.default_dim
        if dtype is None:
            dtype = self.default_dtype
        if mapping is None:
            mapping = self.default_drange_map
        if scale is None:
            scale = self.defualt_drange_scale

        self.unit = unit
        self.dtype = dtype
        self.dim = dim
        self.drange = drange
        self.drange_map = mapping
        self.drange_scale = scale
        self.strict = strict
        self.custom_unit = None



        if default is None:
            if self.default_value is None:
                default = self.get_default()
            else:
                default = self.default_value
        self.default = default

        # initiate the data with default
        if data is None:
            data = self.default
        self._name = ''
        self.data = data
        self._rules = self.default_conversion_rules

    @property
    def unit(self):
        return self._unit
    @unit.setter
    def unit(self, val: str):
        if val in RESERVED:
            uval = unyt.unyt_quantity.from_string('dimensionless')
        else:
            try:
                uval = unyt.unyt_quantity.from_string(val)
                default_unit = unyt.unyt_quantity.from_string(self.default_unit)
                if not default_unit.units == unyt.dimensionless:  # if dimensionless, accept all
                    if not default_unit.units.dimensions.simplify() == uval.units.dimensions.simplify():
                        # if the default is not dimensionless, then the new unit should match the default dimensions
                        raise InvalidUnitChangeError(default_unit.units, default_unit.units.dimensions,
                                                     uval.units, uval.units.dimensions)
            except (UnitParseError, ValueError):
                self.custom_unit = val
                uval = unyt.unyt_quantity.from_string('dimensionless')
        self._name = str(uval.units)
        self._name_dim = str(uval.units.dimensions.simplify())
        self._unit = uval

    @property
    def name(self):
        return self._name

    @property
    def name_dimension(self):
        return self._name_dim

    @property
    def dim(self):
        return self._dim

    @dim.setter
    def dim(self, v):
        self._dim = v

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, v):
        self._default = self.enforce_type(v)

    @property
    def drange(self):
        return self._drange

    @drange.setter
    def drange(self, val):
        drange = []
        for v in val:
            if v is str:  # is str, put it to unyt
                uv = unyt.unyt_quantity.from_string(v)
                if uv.units == unyt.dimensionless:  # in case dimensionless
                    uv *= self.unit
            else:  # otherwise, assume the same unit
                uv = self.enforce_unit(v)
            drange.append(uv)
        self._drange = tuple(drange)

    @property
    def drange_map(self):
        return self._drange_map

    @drange_map.setter
    def drange_map(self, val):
        if val is None:
            self._drange_map = None
            return
        drange = []
        for v in val:
            if callable(v) or None:
                drange.append(v)
                continue
            if isinstance(v, str):  # is str, put it to unyt
                uv = unyt.unyt_quantity.from_string(v)
                if uv.units == unyt.dimensionless:  # in case dimensionless
                    uv *= self.unit
            else:  # otherwise, assume the same unit
                uv = self.enforce_unit(v)
            drange.append(uv)
        self._drange_map = tuple(drange)

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, v):
        self._dtype = v

    @property
    def drange_scale(self):
        return self._drange_scale

    @drange_scale.setter
    def drange_scale(self, v):
        self._drange_scale = v

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, val):
        self._data = self.to(val)     # store date in mutable


    ######################

    def get_default(self):
        if self._is_defdict(self.dtype):
            default = self.dtype.dict()
        elif self.dtype is np.array:
            default = np.zeros(self.dim)
        else:
            try:
                default = self.dtype()
            except (TypeError, AttributeError):
                default = 0.0
        return default

    def enforce_unit(self, val):
        return self._to_unyt(val, self.unit)  # unyt

    def _to_unyt(self, val, unit):
        if val is None:
            return val
        if isinstance(val, str):
            val = unyt.unyt_quantity.from_string(val)
        if not isinstance(unit, str):
            if isinstance(val, unyt.unyt_array):
                val = val.to(unit)
            else:
                val = val * unit
        return val  # unyt

    def enforce(self, val, vdef=None):
        vdef = self._create_class_instance(vdef)
        if vdef.unit.units == self.unit.units:
            return self.enforce_type(val)
        elif self.unit.units == unyt.percent:
            val = vdef.to_percent(val, vdef=self)
        elif self.unit.units == unyt.count:
            val = self.to_count(val, vdef=vdef)
        elif vdef.unit.units == unyt.count:
            val = vdef.from_count(val)
        elif vdef.unit.units == unyt.percent:
            val = self.from_percent(val, vdef=vdef)

        # ruled = self._ruled_conversion(vdef.unit, val, self.unit)
        # if ruled is not None:
        #     return ruled    # if ruled is not None, then return ruled value

        elif not isinstance(val, unyt.unyt_array):
            val = vdef.enforce_unit(val)  # unyt
            val = val.to(self.unit)
        else:
            val = val.to(self.unit)
        return self.enforce_type(val)  # unyt to proper type

    def to(self, val, vdef=None):
        vdef = self._create_class_instance(vdef)
        if vdef.unit.units == self.unit.units:
            return self.enforce_type(val)

        elif vdef.unit.units == unyt.percent:
            val = self.to_percent(val, vdef=vdef)

        elif vdef.unit.units == unyt.count:
            val = vdef.to_count(val, vdef)
        elif self.unit.units == unyt.count:
            val = self.from_count(val)

        elif self.unit.units == unyt.percent:
            val = vdef.from_percent(val, vdef=self)

        # ruled = self._ruled_conversion(vdef.unit, val, self.unit)
        # if ruled is not None:
        #     return ruled    # if ruled is not None, then return ruled value

        elif not isinstance(val, unyt.unyt_array):
            val = self.enforce_unit(val)    # unyt
            val = val.to(vdef.unit)
        else:
            val = val.to(vdef.unit)
        return vdef.enforce_type(val) # unyt to proper type

    def _create_class_instance(self, vdef):
        if vdef is None:
            vdef = self
        elif not isinstance(vdef, UnitType):
            vdef = vdef()
        return vdef

    def from_count(self, val):
        """
        drange_unit should be like
        = (val_0, val_1, ...,  val_n)
        or
        = (val_0, func, val_1, val_2, ..., val_n)
        if you skip func, then linear interpolation will be used
        :param val:
        :return:
        """
        scaled_val = self.enforce_unit(val)
        if self.drange_map is None:
            return scaled_val
        drange_map = iter(self.drange_map)
        drange = iter(self.drange)
        sv_map = next(drange_map)
        sv = next(drange)

        while True:
            try:
                nv_map = next(drange_map)
                lv = next(drange)
                if sv <= val and val <= lv:
                    if not callable(nv_map):  # if not callable not function -> this is the next value
                        lv_map = nv_map
                        func = lambda o_val, sv, lv, sv_map, lv_map: ( ((lv_map - sv_map) / (lv - sv)) * (o_val - sv) + sv_map)
                    else:
                        func = nv_map
                        lv_map = next(drange_map)  # if callable, then we need to get next value

                    scaled_val = func(scaled_val, sv, lv, sv_map, lv_map)
                    break
                else:
                    lv_map = nv_map
            except StopIteration:
                break
            sv_map = lv_map
            sv = lv
        return scaled_val

    def to_count(self, val, vdef=None):
        """
        drange_unit should be like
        = (val_0, val_1, ...,  val_n)
        or
        = (val_0, func, val_1, val_2, ..., val_n)
        if you skip func, then linear interpolation will be used
        :param val:
        :return:
        """
        if vdef is None:
            vdef = self
        scaled_val = vdef._to_unyt(val, vdef.unit)
        if self.drange_map is None:
            return scaled_val
        drange_map = iter(self.drange_map)
        drange = iter(self.drange)
        sv_map = next(drange_map)
        sv = next(drange)

        while True:
            try:
                nv_map = next(drange_map)
                lv = next(drange)
                if sv_map  <= val and val <= nv_map:
                    if not callable(nv_map):  # if not callable not function -> this is the next value
                        lv_map = nv_map
                        func = lambda o_val, sv, lv, sv_map, lv_map: (((lv - sv)/(lv_map - sv_map)) * (o_val - sv_map) + sv)
                    else:
                        func = nv_map
                        lv_map = next(drange_map)  # if callable, then we need to get next value

                    scaled_val = func(scaled_val, sv, lv, sv_map, lv_map)
                    break
                else:
                    lv_map = nv_map
            except StopIteration:
                break
            sv_map = lv_map
            sv = lv
        return scaled_val


    def to_percent(self, val, vdef=None):
        if vdef is None:
            vdef = self
        drange = self.drange
        if all([vd == float('inf') or vd == -float('inf') for vd in drange]):
            drange = vdef.drange
            if all([vd == float('inf') or vd == -float('inf') for vd in drange]):
                raise ValueError(f'{val} cannot converted to % because data range is {self.drange}')

        if not isinstance(val, unyt.unyt_array):
            val = self.enforce_unit(val)
        p_val = (val - drange[MIN]) / (drange[MAX] - drange[MIN]) # percent 0-1 scale
        s_val = p_val * (vdef.drange_scale[MAX] - vdef.drange_scale[MIN]) + vdef.drange_scale[MIN] # scale back to as defined
        return vdef.enforce_type(100*s_val) # percent)

    def from_percent(self, val, vdef=None):
        if vdef is None:
            vdef = self
        drange = self.drange
        if all([vd == float('inf') or vd == -float('inf') for vd in drange]):
            drange = vdef.drange
            if all([vd == float('inf') or vd == -float('inf') for vd in drange]):
                raise ValueError(f'{val} cannot converted to % because data range is {self.drange}')

        val = val/100
        #scale percent i.e. [-100%, 100%], [0,100%]
        p_val = (val - vdef.drange_scale[MIN]) / (vdef.drange_scale[MAX] - vdef.drange_scale[MIN])
        s_val = p_val * (drange[MAX] - drange[MIN]) + drange[MIN]
        return self.enforce_unit(s_val)

    def enforce_type(self, val):
        """
        This will enforce the type and strip off the unit if unyt variables are given
        """
        enforced_val = val
        if val is None:
            return

        if self._is_defdict(self.dtype):
            self.dtype.set(val)
            return self.dtype

        if self.dtype is np.array:  # special handling for np.array
            if isinstance(val, unyt.unyt_array):
                enforced_val = val.to_ndarray()
            elif not isinstance(val, self.dtype):
                enforced_val = np.array(val)
        else:
            if isinstance(self.dtype, type) and not isinstance(val, self.dtype): # otherwise apply the type
                try:
                    enforced_val = self.dtype(val)
                except (TypeError, AttributeError): # some type like Any throw the error
                    pass
        try:
            enforced_val = np.clip(enforced_val, *tuple([d.value for d in self.drange]))
        except TypeError:
            pass
        return enforced_val

    def _ruled_conversion(self, from_unit, val, to_unit):
        if self._rules:
            for rule in self._rules:
                if rule.if_rule_apply(from_unit) and rule.if_rule_out_contains(to_unit):
                    return rule.bind(dict(from_unit=val))
                if rule.if_rule_apply(to_unit) and rule.if_rule_out_contains(from_unit):
                    return rule.inv_bind(to_unit)
        return None

    def _is_defdict(self, data):
        try:
            if data.is_DefDict():
                return True
        except AttributeError:
            pass
        return False



    def __str__(self):
        return f"{self.unit.units, tuple([d.to_string() for d in self.drange]), self.dtype}"

    def __repr__(self):
        return  f"{self.unit.units, tuple([d.to_string() for d in self.drange]), self.dtype}"





from setuptools import setup, find_packages
import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_metadata(field):
    rel_path = "defdict/__init__.py"
    for line in read(rel_path).splitlines():
        if line.startswith(f'__{field}__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError(f"Unable to find {field} string.")

# Loads _version.py module without importing the whole package.
def get_version_and_cmdclass(pkg_path):
    from importlib.util import module_from_spec, spec_from_file_location
    spec = spec_from_file_location(
        'version', os.path.join(pkg_path, '_version.py'),
    )
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.get_cmdclass(pkg_path)

version, cmdclass = get_version_and_cmdclass('defdict')


with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='defdict',
    version='0.0.2',
    cmdclass=cmdclass,
    description=get_metadata('description'),
    url=get_metadata("url"),
    author=get_metadata('author'),
    license='LGPLv3',
    packages=find_packages(include=['defdict', 'defdict.*']),
    install_requires=requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Robot Framework',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3',
    ],
)

